// RoffParameterDialog.cpp : implementation file
//

#include "stdafx.h"
#include "RoffView.h"
#include "RoffParameterDialog.h"
#include "RoffPhongFactorDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// RoffParameterDialog dialog


RoffParameterDialog::RoffParameterDialog(CWnd* pParent /*=NULL*/)
	: CDialog(RoffParameterDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(RoffParameterDialog)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
	dialogTransformer.setAt(RoffVector(0,0,0));
	dialogTransformer.setFrom(RoffVector(0,0,-2.5));
	dialogTransformer.setUpVector(RoffVector(0,1.0,0));
	dialogTransformer.setAngle(45.0);
}

void RoffParameterDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(RoffParameterDialog)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(RoffParameterDialog, CDialog)
	//{{AFX_MSG_MAP(RoffParameterDialog)
	ON_BN_CLICKED(IDC_EXT_COLOR, OnExtColor)
	ON_BN_CLICKED(IDC_INT_COLOR, OnIntColor)
	ON_BN_CLICKED(IDC_HIGH_COLOR, OnHighColor)
	ON_BN_CLICKED(IDC_ALIGHT, OnAlight)
	ON_BN_CLICKED(IDC_DLIGHT, OnDlight)
	ON_BN_CLICKED(ID_PD_RESET, OnPdReset)
	ON_BN_CLICKED(ID_PD_ADVANCED, OnAdvancedSettings)
	ON_WM_HSCROLL()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// RoffParameterDialog message handlers
BOOL RoffParameterDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	CSliderCtrl* slider = (CSliderCtrl*) GetDlgItem(IDC_SHINY);
	slider->SetRangeMin(1);
	slider->SetRangeMax(30);

	slider = (CSliderCtrl*) GetDlgItem(IDC_XAXIS);
	slider->SetRangeMin(0);
	slider->SetRangeMax(20);

	slider = (CSliderCtrl*) GetDlgItem(IDC_YAXIS);
	slider->SetRangeMin(0);
	slider->SetRangeMax(20);

	slider = (CSliderCtrl*) GetDlgItem(IDC_ZAXIS);
	slider->SetRangeMin(0);
	slider->SetRangeMax(20);

    setAllValues();
    UpdatePreview();
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void RoffParameterDialog::setIlluminator(RoffPhong illuminator)
{

	dialogIlluminator = illuminator;
}

RoffPhong RoffParameterDialog::getIlluminator()
{
	return dialogIlluminator;
}

void RoffParameterDialog::setAllValues()
{
	CMetaFileDC newMetaFile;

	// Exterior color
	updateColorBox(IDC_EXT_COLOR, dialogIlluminator.getOutsideColor());

	// Interior color
	updateColorBox(IDC_INT_COLOR, dialogIlluminator.getInsideColor());

	// Highlight color
	updateColorBox(IDC_HIGH_COLOR, dialogIlluminator.getHighlightColor());

	// Ambient Light color
	updateColorBox(IDC_ALIGHT, dialogIlluminator.getAmbientColor());

	// Directional Light color
	updateColorBox(IDC_DLIGHT, dialogIlluminator.getDirectionalColor());

	// Shininess factor
	CSliderCtrl* slider = (CSliderCtrl*) GetDlgItem(IDC_SHINY);
    slider->SetPos((int)dialogIlluminator.getShininess());

	CSliderCtrl* xSlider = (CSliderCtrl*) GetDlgItem(IDC_XAXIS);
	CSliderCtrl* ySlider = (CSliderCtrl*) GetDlgItem(IDC_YAXIS);
	CSliderCtrl* zSlider = (CSliderCtrl*) GetDlgItem(IDC_ZAXIS);
	RoffVector light = dialogIlluminator.getLightDirection();
    light = light / light.getLength();
    xSlider->SetPos((int)(light.x * 10.0) + 10);
    ySlider->SetPos((int)(light.y * 10.0) + 10);
    zSlider->SetPos((int)(light.z * 10.0) + 10);
}

void RoffParameterDialog::OnExtColor() 
{
	CColorDialog picker(dialogIlluminator.getOutsideColor());
	COLORREF newColor;

	if (picker.DoModal() == IDOK) {
		newColor = picker.GetColor();
		dialogIlluminator.setOutsideColor(newColor);
		updateColorBox(IDC_EXT_COLOR, newColor);
		UpdatePreview();
	}	
}

void RoffParameterDialog::OnIntColor() 
{
	CColorDialog picker(dialogIlluminator.getInsideColor());
	COLORREF newColor;

	if (picker.DoModal() == IDOK) {
		newColor = picker.GetColor();
		dialogIlluminator.setInsideColor(newColor);
		updateColorBox(IDC_INT_COLOR, newColor);
		UpdatePreview();
	}	
}

void RoffParameterDialog::OnHighColor() 
{
	CColorDialog picker(dialogIlluminator.getHighlightColor());
	COLORREF newColor;

	if (picker.DoModal() == IDOK) {
		newColor = picker.GetColor();
		dialogIlluminator.setHighlightColor(newColor);
		updateColorBox(IDC_HIGH_COLOR, newColor);
		UpdatePreview();
	}	
}

void RoffParameterDialog::OnAlight() 
{
	CColorDialog picker(dialogIlluminator.getAmbientColor());
	COLORREF newColor;

	if (picker.DoModal() == IDOK) {
		newColor = picker.GetColor();
		dialogIlluminator.setAmbientColor(newColor);
		updateColorBox(IDC_ALIGHT, newColor);
		UpdatePreview();
	}	
}

void RoffParameterDialog::OnDlight() 
{
	CColorDialog picker(dialogIlluminator.getDirectionalColor());
	COLORREF newColor;

	if (picker.DoModal() == IDOK) {
		newColor = picker.GetColor();
		dialogIlluminator.setDirectionalColor(newColor);
		updateColorBox(IDC_DLIGHT, newColor);
		UpdatePreview();
	}	
}

void RoffParameterDialog::OnPdReset() 
{
	dialogIlluminator = RoffPhong();
	setAllValues();
	UpdatePreview();
}

void RoffParameterDialog::OnAdvancedSettings() 
{
	RoffPhongFactorDialog rpfd;
    rpfd.m_ambient  = dialogIlluminator.getAmbientFactor();
	rpfd.m_diff     = dialogIlluminator.getDiffuseFactor();
	rpfd.m_specular = dialogIlluminator.getSpecularFactor();

	if (rpfd.DoModal() == IDOK) {
		// Get the values back out
		dialogIlluminator.setAmbientFactor(rpfd.m_ambient);
		dialogIlluminator.setDiffuseFactor(rpfd.m_diff);
		dialogIlluminator.setSpecularFactor(rpfd.m_specular);
		UpdatePreview();
	}
}

void RoffParameterDialog::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
	bool redraw = false;

	CSliderCtrl* shinySlider = (CSliderCtrl*) GetDlgItem(IDC_SHINY);
	CSliderCtrl* xSlider = (CSliderCtrl*) GetDlgItem(IDC_XAXIS);
	CSliderCtrl* ySlider = (CSliderCtrl*) GetDlgItem(IDC_YAXIS);
	CSliderCtrl* zSlider = (CSliderCtrl*) GetDlgItem(IDC_ZAXIS);

	if (shinySlider == (CSliderCtrl*) pScrollBar) {
		dialogIlluminator.setShininess((double)shinySlider->GetPos());
		redraw = true;
	} else if ((xSlider == (CSliderCtrl*) pScrollBar) ||
		       (ySlider == (CSliderCtrl*) pScrollBar) || 
			   (zSlider == (CSliderCtrl*) pScrollBar)) {
		RoffVector newLight = RoffVector((xSlider->GetPos() - 10) * 10.0,
			                             (ySlider->GetPos() - 10) * 10.0,
										 (zSlider->GetPos() - 10) * 10.0);
		if (newLight.getLength() == 0) {
			newLight = RoffVector(0, 0, -1);
		} else {
		    newLight = newLight / newLight.getLength();
		}
		dialogIlluminator.setLightDirection(newLight);
        xSlider->SetPos((int)(newLight.x * 10.0) + 10);
        ySlider->SetPos((int)(newLight.y * 10.0) + 10);
        zSlider->SetPos((int)(newLight.z * 10.0) + 10);
		redraw = true;
	}
	if (redraw) {
		UpdatePreview();
	}
}

// This function creates a ray-cast sphere in the preview
// panel, using the current lighting settings.
void RoffParameterDialog::UpdatePreview() {

	CStatic* viewPane = (CStatic*) GetDlgItem(IDC_PD_VIEWPANE);
	COLORREF drawColor;

	CBitmap* bitmap = new CBitmap();
	CDC dcMemory;
	CDC* windowDC = GetDC();
	// The viewPane is 128 x 108
	bitmap->CreateCompatibleBitmap(windowDC, 128, 108);
	dcMemory.CreateCompatibleDC(windowDC);
	ReleaseDC(windowDC);
	CBitmap* pOldBitmap = dcMemory.SelectObject(bitmap);
	
	int pixelsX = 128;
	int pixelsY = 108;
	int halfPixelsX = 64;
	int halfPixelsY = 54;

	// Variables for temporary storage during processing.
	double u;
	double v;
	double a;
	double t1;
	double t2;
	double det;
	RoffVector t1Point;
	RoffVector t2Point;
    RoffVector normal;
	RoffVector view;

	// Set the illumination.
	dialogIlluminator.setWorldLight(dialogTransformer.EyeToWorld(
		                              dialogIlluminator.getLightDirection()));

	dcMemory.FillSolidRect(0,0,pixelsX,pixelsY, 0);
	for (int j = 0; j < pixelsY; j++) {
	    for (int i = 0; i < pixelsX; i++) {
			// Find normalized double values, centered at 0,0, for i,j.
			// Use the X size, as it's the larger of the two.
			u = (i - halfPixelsX) / (float)pixelsX;
            v = (j - halfPixelsY) / (float)pixelsX;
			a = (u * u + v * v + 1.0);
			// det = b^2 - 4 * a * c
			det = 25.0 - 21.0 * a;
			// If there are roots, then each intersection
			// is where the ray hits the sphere.
			if (det > 0) {
				// Find both intersections.
				t1 = (5.0 + sqrt(det)) / (2.0 * a);
				t2 = (5.0 - sqrt(det)) / (2.0 * a);
				t1Point.x = -u * t1;
				t1Point.y = v * t1;
				t1Point.z = t1 - 2.5;
				t2Point.x = -u * t2;
				t2Point.y = v * t2;
				t2Point.z = t2 - 2.5;
				// Note: Because t1Point and t2Point are
				// points on a unit sphere, their lengths are
				// both 1, and they are the normal of the sphere
				// at their respective points.
				// Select the lower z of the two, which is the closer point.
                if (t1Point.z <= 0) {
					view   = dialogTransformer.getFrom() - t1Point;
					view   = view / view.getLength();
					if (t1Point.Dot(view) > 0) {
						drawColor = dialogIlluminator.Illuminate(t1Point,
		                                                         view);
					} else {
						drawColor = RGB(0,0,0);
					}
				} else if (t2Point.z <= 0) {
					view   = dialogTransformer.getFrom() - t2Point;
					view   = view / view.getLength();
					if (t2Point.Dot(view) > 0) {
						drawColor = dialogIlluminator.Illuminate(t2Point,
		                                                         view);
					} else {
						drawColor = RGB(0,0,0);
					}
				} else {
					drawColor = RGB(0,0,0);
				}
			} else {
				drawColor = RGB(0,0,0);
			}
			dcMemory.SetPixelV(i, pixelsY - j, drawColor);
		}
	}
	dcMemory.SelectObject(pOldBitmap);
	pOldBitmap = CBitmap::FromHandle(viewPane->SetBitmap(*bitmap));
	if (pOldBitmap) {
		delete pOldBitmap;
	}
	// Make sure it gets re-drawn.
	viewPane->Invalidate(FALSE);
}

void RoffParameterDialog::updateColorBox(int id, COLORREF color)
{
	CStatic* colorBox = (CStatic*) GetDlgItem(id);
	CBitmap* bitmap = new CBitmap();
	CDC dcMemory;
	CDC* windowDC = GetDC();
	// Create a 5x5 bitmap. It will get centered, and the
	// remainder will be filled with the color from
	// its upper left corner.
	bitmap->CreateCompatibleBitmap(windowDC, 5, 5);
	dcMemory.CreateCompatibleDC(windowDC);
	ReleaseDC(windowDC);
	// Load the bitmap into the DC
	CBitmap* pOldBitmap = dcMemory.SelectObject(bitmap);

	// Fill it with color
	dcMemory.FillSolidRect(0, 0, 5, 5, color);

	// Pull it back out, putting the default one back in.
	dcMemory.SelectObject(pOldBitmap);
	// Pass it to the color box.
	pOldBitmap = CBitmap::FromHandle(colorBox->SetBitmap(*bitmap));
	// If the color box had an old one, get rid of it.
	if (pOldBitmap) {
		delete pOldBitmap;
	}
}

void RoffParameterDialog::clearStaticBMP(int id)
{
	CStatic* staticItem = (CStatic*) GetDlgItem(id);
	CBitmap* pOldBitmap = CBitmap::FromHandle(staticItem->SetBitmap(NULL));
	if (pOldBitmap) {
		delete pOldBitmap;
	}
}

void RoffParameterDialog::clearBMPs()
{
	// Exterior color
	clearStaticBMP(IDC_EXT_COLOR);

	// Interior color
	clearStaticBMP(IDC_INT_COLOR);

	// Highlight color
	clearStaticBMP(IDC_HIGH_COLOR);

	// Ambient Light color
	clearStaticBMP(IDC_ALIGHT);

	// Directional Light color
	clearStaticBMP(IDC_DLIGHT);

	// Preview pane
	clearStaticBMP(IDC_PD_VIEWPANE);
}

void RoffParameterDialog::OnOK() 
{
	clearBMPs();
	
	CDialog::OnOK();
}

void RoffParameterDialog::OnCancel() 
{
	clearBMPs();
	
	CDialog::OnCancel();
}
