// RoffPerspective.cpp: implementation of the RoffPerspective class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RoffView.h"
#include "RoffPerspective.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RoffPerspective::RoffPerspective()
{

	upVector = RoffVector(0, 1, 0);
	atPoint = RoffVector(0, 0, 0);
	fromPoint = RoffVector(1, 1, 1);
	viewAngle = 45;
	factor2D = 1.0/(2.0 * tan(viewAngle * 3.14159265358979 / 360.0));
	UpdateTransformation();
}

RoffPerspective::~RoffPerspective()
{

}

void RoffPerspective::setAngle(const double newAngle) {

	viewAngle = newAngle;
	if (viewAngle > 180) {
		viewAngle = 180.0;
	} else if (viewAngle < .5) {
		viewAngle = .5;
	}
	factor2D = 1.0/(2.0 * tan(viewAngle * 3.14159265358979 / 360.0));
}

void RoffPerspective::UpdateTransformation() {
	
	RoffVector numeratorA;
	RoffVector numeratorB;
	RoffVector numeratorC = atPoint - fromPoint;
	double denominator;
	
	// NumeratorC is A - F
	denominator = numeratorC.getLength();

	cColumn = numeratorC / denominator;
	
	// NumeratorA is (A - F) x U
	numeratorA = numeratorC.Cross(upVector);
	denominator = numeratorA.getLength();

	aColumn = numeratorA / denominator;

	// NumeratorB is (A - F) x U x (A - F)
	numeratorB = numeratorA.Cross(numeratorC);
	denominator = numeratorB.getLength();

	bColumn = numeratorB / denominator;
}

void RoffPerspective::zoom(const double percentage) {

	double scaleFactor = percentage / 100.0;
	bool validScale = false;
    RoffVector lineOfSight = atPoint - fromPoint;

	if (percentage < 100) {
		if (lineOfSight.getLength() > .1) {
			validScale = true;
		}
	} else {
		if (lineOfSight.getLength() < 5000) {
			validScale = true;
		}
	}
	if (validScale) {
		fromPoint = ((fromPoint - atPoint) * scaleFactor) + atPoint;
	}

	UpdateTransformation();
}

void RoffPerspective::translate(const double rightPercent, const double upPercent){
	
	RoffVector vectorLoS = atPoint - fromPoint;

	RoffVector horizon = upVector.Cross(vectorLoS);
	RoffVector centerLine = vectorLoS.Cross(horizon);

	RoffVector offsets = (horizon * rightPercent) + (centerLine * upPercent);

	atPoint = atPoint + offsets;
	fromPoint = fromPoint + offsets;
}

void RoffPerspective::rotate(const double upDegrees, const double rightDegrees){

	fromPoint = fromPoint - atPoint;
	RoffVector horizon = fromPoint.Cross(upVector);
	RoffVector centerLine = horizon.Cross(fromPoint);

	if (upDegrees != 0) {
		fromPoint = fromPoint.rotate(horizon, upDegrees);
		upVector = upVector.rotate(horizon, upDegrees);
	}

	if (rightDegrees != 0) {
		fromPoint = fromPoint.rotate(centerLine, rightDegrees);
		upVector = upVector.rotate(centerLine, rightDegrees);
	}

	fromPoint = fromPoint + atPoint;

	UpdateTransformation();
}

RoffVector RoffPerspective::WorldToEye(const RoffVector worldPoint) const {

	// World point - F
	RoffVector factors = worldPoint - fromPoint;
	
	return RoffVector(factors.Dot(aColumn),
		              factors.Dot(bColumn),
					  factors.Dot(cColumn));
}

RoffVector RoffPerspective::EyeToWorld(const RoffVector eyePoint) const {

	// eyePoint * V^T
	return RoffVector((eyePoint.x * aColumn.x +
		                               eyePoint.y * bColumn.x +
									   eyePoint.z * cColumn.x),
                                      (eyePoint.x * aColumn.y +
		                               eyePoint.y * bColumn.y +
									   eyePoint.z * cColumn.y),
									  (eyePoint.x * aColumn.z +
		                               eyePoint.y * bColumn.z +
									   eyePoint.z * cColumn.z));
}

RoffVector RoffPerspective::EyeTo2D(const RoffVector toConvert) const {

	return RoffVector((factor2D * toConvert.x / toConvert.z) + .5, 
		              (factor2D * toConvert.y / toConvert.z) + .5,
					  0);
}
