// RoffVector.cpp: implementation of the RoffVector class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RoffView.h"
#include "RoffVector.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RoffVector::RoffVector()
{
    x = 0;
	y = 0;
	z = 0;
}

RoffVector::RoffVector(double newX, double newY, double newZ)
{
    x = newX;
	y = newY;
	z = newZ;
}

RoffVector::~RoffVector()
{

}

double RoffVector::getLength() const {

	return sqrt(x * x + y * y + z * z);
}

double RoffVector::getSquaredLength() const {

	return (x * x + y * y + z * z);
}

RoffVector RoffVector::Cross(const RoffVector& rhs) const {

	RoffVector rval;

	rval.x = (y * rhs.z) - (z * rhs.y);
	rval.y = (z * rhs.x) - (x * rhs.z);
	rval.z = (x * rhs.y) - (y * rhs.x);

	return rval;
}

double RoffVector::Dot(const RoffVector& rhs) const {

	return (x * rhs.x + y * rhs.y + z * rhs.z);
}

RoffVector RoffVector::operator+(const RoffVector& rhs) const {
    return RoffVector(x + rhs.x, y + rhs.y, z + rhs.z);
}

RoffVector RoffVector::operator-(const RoffVector& rhs) const {
    return RoffVector(x - rhs.x, y - rhs.y, z - rhs.z);
}

RoffVector RoffVector::operator*(const double factor) const {
    return RoffVector(x * factor, y * factor, z * factor);
}

RoffVector RoffVector::operator/(const double factor) const {
    return RoffVector(x / factor, y / factor, z / factor);
}

RoffVector RoffVector::rotate(const RoffVector axis, double degrees) const {

	RoffVector rval;
	double cosAngle = cos(degrees * 3.14159265358979 / 180.0);
	double sinAngle = sin(degrees * 3.14159265358979 / 180.0);
	RoffVector rotatedToZ;
	RoffVector rx;
	RoffVector ry;
	RoffVector rz;

	// Normalize the axis and store in rz for later
	// potential use.
	rz = axis / axis.getLength();

	// If this vector is parallel to the axis (Now in rz),
	// don't bother rotating it. Check with Cauchy-Schwartz
	// u dot v == length of u * length of v iff u is linearly
	// dependant on v. Length of rz is one, because it is
	// normalized.
	if (fabs((*this).Dot(rz)) == (*this).getLength()) {
		rval = *this;
	} else {
	    // If we're not already rotating around the Z, transform to Z
	    if (axis.x == 0 && axis.y == 0) {
            // In this case, the axis is along Z already, so we
			// wont bother rotating our axis of rotation to Z.
			rotatedToZ = *this;
		} else {
	        // Build the rotation matrix
	        // rz was assigned already while normalizing the axis.
	        rx = (*this).Cross(axis);
	        rx = rx / rx.getLength();
	        ry = rz.Cross(rx);

	        // Move this vector such that the axis would be in Z
	        rotatedToZ = RoffVector(rx.Dot(*this), ry.Dot(*this), rz.Dot(*this));
		}

	    // Rotate this vector around Z
	    RoffVector rotatedAroundZ = 
		RoffVector(rotatedToZ.x * cosAngle - rotatedToZ.y * (-1.0) * sinAngle,
		           rotatedToZ.x * sinAngle + rotatedToZ.y * cosAngle,
				   rotatedToZ.z);
		                                   
	    if (axis.x == 0 && axis.y == 0) {
		    rval = rotatedAroundZ;
		} else {
    	    // Move back so axis is in original location.
	        rval.x = rotatedAroundZ.x * rx.x + 
		             rotatedAroundZ.y * ry.x +
			         rotatedAroundZ.z * rz.x;
	        rval.y = rotatedAroundZ.x * rx.y + 
		             rotatedAroundZ.y * ry.y +
			         rotatedAroundZ.z * rz.y;
	        rval.z = rotatedAroundZ.x * rx.z + 
		             rotatedAroundZ.y * ry.z +
		    	     rotatedAroundZ.z * rz.z;
		}
	}

    return rval;
}