// RoffViewView.cpp : implementation of the CRoffViewView class
//

#include "stdafx.h"
#include "RoffView.h"

#include "RoffViewDoc.h"
#include "RoffViewView.h"
#include "RoffModelInfoDialog.h"
#include "RoffParameterDialog.h"
#include "RoffViewParameterDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CRoffViewView

IMPLEMENT_DYNCREATE(CRoffViewView, CView)

BEGIN_MESSAGE_MAP(CRoffViewView, CView)
	//{{AFX_MSG_MAP(CRoffViewView)
	ON_WM_SIZE()
	ON_COMMAND(ID_TOOLS_BGCOLOR, OnChangeBGColor)
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_COMMAND(ID_VIEW_MODEL_INFO, OnViewModelInfo)
	ON_COMMAND(ID_TOOLS_RESETVIEW, OnToolsResetview)
	ON_COMMAND(ID_TOOLS_PARAMS, OnParameterDialog)
	ON_COMMAND(ID_VIEW_PARAMETERS, OnViewParameters)
	ON_COMMAND(ID_VIEW_SOLID, OnViewSolid)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SOLID, OnUpdateViewSolid)
	ON_COMMAND(ID_VIEW_WIRE, OnViewWire)
	ON_UPDATE_COMMAND_UI(ID_VIEW_WIRE, OnUpdateViewWire)
	ON_UPDATE_COMMAND_UI(ID_VIEW_POINTS, OnUpdateViewPoints)
	ON_COMMAND(ID_VIEW_POINTS, OnViewPoints)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_CONTROL_EDIT, OnControlEdit)
	ON_UPDATE_COMMAND_UI(ID_CONTROL_EDIT, OnUpdateControlEdit)
	ON_COMMAND(ID_CONTROL_VIEW, OnControlView)
	ON_UPDATE_COMMAND_UI(ID_CONTROL_VIEW, OnUpdateControlView)
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CRoffViewView construction/destruction

CRoffViewView::CRoffViewView()
{
	pMemDC        = NULL;
	bgColor       = RGB(0, 0, 0);
	drawPolygons  = false;
	drawPoints    = false;
	drawWireframe = true;
	numPoints     = 0;
	lmouseDown    = false;
	rmouseDown    = false;
	viewMode      = true;
	controlMode   = false;
	selectedPoint = -1;
}

CRoffViewView::~CRoffViewView()
{
	// Delete the screen buffer.
    if (pMemDC != NULL) {
	    DeleteObject(pMemDC->GetCurrentBitmap());
	    pMemDC->DeleteDC();
	}
	delete pMemDC;
}

BOOL CRoffViewView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CRoffViewView drawing

void CRoffViewView::OnDraw(CDC* pDC)
{
	// BitBlt from the buffer to the screen.
	pDC->BitBlt(upperX, upperY, squareSize, squareSize, pMemDC, 0, 0, SRCCOPY);
}

/////////////////////////////////////////////////////////////////////////////
// CRoffViewView diagnostics

#ifdef _DEBUG
void CRoffViewView::AssertValid() const
{
	CView::AssertValid();
}

void CRoffViewView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CRoffViewDoc* CRoffViewView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CRoffViewDoc)));
	return (CRoffViewDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CRoffViewView message handlers

void CRoffViewView::BuildBuffer()
{
    CRoffViewDoc* pDoc = GetDocument();
		
	// Fill the background with the background color
	pMemDC->FillSolidRect(0,0,squareSize,squareSize, bgColor);
	
	// Build and select a new pen in the model color
	CPen lineColor;
	lineColor.CreatePen(0,0, illuminator.getOutsideColor());
	CPen* oldPen = pMemDC->SelectObject(&lineColor);
	// Build and select a new brush in the background color
	CBrush fillBrush;
	fillBrush.CreateSolidBrush(bgColor);
	CBrush* oldBrush = pMemDC->SelectObject(&fillBrush);

	if (drawWireframe) {
		for (int i = 0; i < numPoints;i += 2) {
			pMemDC->MoveTo((int)(convertedXPoints[pDoc->getPointIndex(i)] * squareSize),
				           squareSize - (int)(convertedYPoints[pDoc->getPointIndex(i)] * squareSize));
			pMemDC->LineTo((int)(convertedXPoints[pDoc->getPointIndex(i + 1)] * squareSize),
				           squareSize - (int)(convertedYPoints[pDoc->getPointIndex(i + 1)] * squareSize));
		}
	} else if (drawPolygons) {
		int numTriangles = tmanager.getNumTriangles();
		COLORREF polyColor;
		RoffTriangle* tri;
		CPoint* trianglePoints = NULL;
		tmanager.setScreenFactor(squareSize);
	    illuminator.setWorldLight(transformer.EyeToWorld(illuminator.getLightDirection()));
		for (int i = 0;i < numTriangles;i++){
			tri = tmanager.get(i);
			// Get the illumination for this particular triangle.
			polyColor = illuminator.Illuminate(tri->getNormal(), tri->getView());
			// Select in pen and brush of this color.
			pMemDC->SelectObject(oldPen);
			lineColor.DeleteObject();
			lineColor.CreatePen(0,0, polyColor);
			oldPen = pMemDC->SelectObject(&lineColor);
			pMemDC->SelectObject(oldBrush);
			fillBrush.DeleteObject();
			fillBrush.CreateSolidBrush(polyColor);
			oldBrush = pMemDC->SelectObject(&fillBrush);
			// Get the points of the triangle.
			trianglePoints = tmanager.getCPoints(i);
			// Draw the triangle
			pMemDC->Polygon(trianglePoints, 3);
			delete trianglePoints;
	    }
	} else if (drawPoints) {
		// Draw all the points in the array.
		int totalVertices = pDoc->numVertices;
		COLORREF pointColor = illuminator.getOutsideColor();
		for (int i = 0;i < totalVertices;i++) {
            pMemDC->SetPixelV((int)(convertedXPoints[i] * squareSize),
				              (int)(squareSize - (convertedYPoints[i] * squareSize)),
							  pointColor);
		}
	}
	// If a point is selected, put a box around it.
	if (selectedPoint > -1) {
		int cornerX = (int)(convertedXPoints[selectedPoint] * squareSize) - 2;
		int cornerY = squareSize - (int)(convertedYPoints[selectedPoint] * squareSize) - 2;
		pMemDC->Draw3dRect(cornerX, cornerY, 4, 4, RGB(170, 170, 170), RGB(255, 255,255));
	}

	// Clean up
	pMemDC->SelectObject(oldPen);
	pMemDC->SelectObject(oldBrush);
}

void CRoffViewView::BuildStructures()
{
	CRoffViewDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	RoffVector point2D;

	// Convert the 3D points to 2D points.
	int totalVertices = pDoc->numVertices;
	for (int i = 0;i < totalVertices;i++) {
        point2D = transformer.EyeTo2D(transformer.WorldToEye(pDoc->vertices[i]));
        convertedXPoints[i] = point2D.x;
	    convertedYPoints[i] = point2D.y;
	}
	// Depth-sort the triangles based on the viewer
	// location, if we are shading the triangles.
	if (drawPolygons) {
		tmanager.orderTriangles(transformer.getFrom());
	}
}

void CRoffViewView::SetDefaultPerspective()
{
	CRoffViewDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// Calculate the largest span of the bounding
	// box of the current object.
	double minX  = pDoc->minX;
	double maxX  = pDoc->maxX;
	double minY  = pDoc->minY;
	double maxY  = pDoc->maxY;
	double minZ  = pDoc->minZ;
	double maxZ  = pDoc->maxZ;

	double height = maxY - minY;
	double width  = maxX - minX;
	double depth  = maxZ - minZ;
	double maxDim = height;
	if (width > maxDim) {
		maxDim = width;
	}
	if (depth > maxDim) {
		maxDim = depth;
	}

	// Look at the center of the bounding box.
	RoffVector newAt = RoffVector((maxX + minX)/2.0,
		                          (maxY + minY)/2.0,
					              (maxZ + minZ)/2.0);
	transformer.setAt(newAt);
	// View from along the Z-axis, 2.5 times the largest
	// dimension of the object away. This insures a
	// reasonable view of the object.
	newAt.z = newAt.z + (maxDim * 2.5);
	transformer.setFrom(newAt);
	transformer.setUpVector(RoffVector(0, 1.0, 0));
	transformer.setAngle(45);
}

void CRoffViewView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);

	CRect rect;
	GetClientRect(rect);
	if (rect.right > rect.bottom) {
		upperY = rect.top;
		upperX = (rect.right / 2) - (rect.bottom / 2);
		squareSize = rect.bottom;
	} else {
		upperX = rect.left;
		upperY = (rect.bottom / 2) - (rect.right / 2);
		squareSize = rect.right;
	}

	// Create the buffer if it doesn't exist.
	if (pMemDC == NULL) {
		pMemDC = new CDC;
		pMemDC->CreateCompatibleDC(NULL);
	}
	CBitmap newBitmap;
	// Get the device context to match the bitmap with.
	CDC* windowDC = GetDC();
	// Create a new square bitmap of the size just calculated
	// for the memory buffer.
	newBitmap.CreateCompatibleBitmap(windowDC, squareSize, squareSize);
	// Release the DC, now that we've used it.
	ReleaseDC(windowDC);
	// Select the bitmap into the buffer.
	DeleteObject(pMemDC->SelectObject(&newBitmap));

	pMemDC->SetWindowOrg(0, 0);

	BuildBuffer();
}

void CRoffViewView::OnChangeBGColor() 
{
	CColorDialog picker(bgColor);

	if (picker.DoModal() == IDOK) {
		bgColor = picker.GetColor();
		BuildBuffer();
		Invalidate(FALSE);
	}	
}

void CRoffViewView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	CView::OnLButtonDown(nFlags, point);
	lmouseDown = true;
	lastClickPoint = point;
}

void CRoffViewView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	CView::OnLButtonUp(nFlags, point);
	lmouseDown = false;
}

void CRoffViewView::OnMouseMove(UINT nFlags, CPoint point) 
{
	CView::OnMouseMove(nFlags, point);

	bool redraw = false;

	// If we are viewing: i.e. Rotating, zooming or shift-panning.
	if (viewMode) {
        // If both buttons down,
	    // shift pan along current "axes".
	    if (rmouseDown && lmouseDown) {
		    double xChange = 0;
		    double yChange = 0;

		    if (point.x < lastClickPoint.x - 3) {
			    xChange = -.05;
		    } else if (point.x > lastClickPoint.x + 3) {
			    xChange = .05;
		    }
		    if (point.y < lastClickPoint.y - 3) {
    			yChange = -.05;
	    	} else if (point.y > lastClickPoint.y + 3) {
		    	yChange = .05;
		    }
		    transformer.translate(xChange, yChange);
		    redraw = true;
	    // If only the right button, zoom in and out by 5%.
	    } else if (rmouseDown) {
			if (point.y < lastClickPoint.y) {
				// Zoom in
	    		transformer.zoom(95.0);
				redraw = true;
		    } else if (point.y > lastClickPoint.y) {
				// Zoom out
				transformer.zoom(105.0);
			    redraw = true;
		    }
	    // If only the left button, rotate around At point.
	    } else if (lmouseDown) {
    		double rightDegrees = 0;
	    	double upDegrees = 0;
		
		    if (point.x < lastClickPoint.x - 1) {
			    rightDegrees = (double)(point.x - lastClickPoint.x);
    		} else if (point.x > lastClickPoint.x + 1) {
	    		rightDegrees = (double)(point.x - lastClickPoint.x);
		    }
		    if (point.y < lastClickPoint.y - 1) {
    			upDegrees = (double) (lastClickPoint.y - point.y);
	    	} else if (point.y > lastClickPoint.y + 1) {
		    	upDegrees = (double) (lastClickPoint.y - point.y);
		    }
		    transformer.rotate(upDegrees, rightDegrees);
		    redraw = true;
	    }
	// If we're not viewing, we're in control mode. If we have
	// have selected a point, we need to move the point around.
	} else if (controlMode && (selectedPoint != -1)) {
		CRoffViewDoc* pDoc = GetDocument();
		int xDiff = point.x - lastClickPoint.x; // Positive indicates rightward movement.
		int yDiff = point.y - lastClickPoint.y; // Positive indicates downward movement.
		RoffVector towardFrom = transformer.getFrom() - transformer.getAt(); // Positive is toward from
       	RoffVector horizon = towardFrom.Cross(transformer.getUp()); // Positive is left.
        RoffVector centerLine = horizon.Cross(towardFrom); // Positive is up
		RoffVector change; // Holds the delta vector to change the point's position by.

		if (lmouseDown) {
			// Drag left or right and/or up or down.
			change = horizon * (xDiff * -.01);
			change = change + (centerLine * yDiff * -.01);
            pDoc->vertices[selectedPoint] = pDoc->vertices[selectedPoint] + change;
			redraw = true;
		} else if (rmouseDown) {
			// Drag in and out.
            change = towardFrom * (yDiff * .01);
            pDoc->vertices[selectedPoint] = pDoc->vertices[selectedPoint] + change;
			redraw = true;
		}
		// Recalculate the normals and centroids for the triangles.
		if (redraw) {
		    tmanager.rebuild();
		}
	}

	// Only do the work of re-building things if something happened
	// that makes it necessary.
	if (redraw) {
		BuildStructures();
		BuildBuffer();
		Invalidate(FALSE);
	}

	// Save the current point for the next call.
	lastClickPoint = point;
}

void CRoffViewView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	CView::OnRButtonDown(nFlags, point);
	rmouseDown = true;
	lastClickPoint = point;
}

void CRoffViewView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	CView::OnRButtonUp(nFlags, point);
	rmouseDown = false;
}

void CRoffViewView::OnViewModelInfo() 
{
	CRoffModelInfoDialog theDialog;
	CRoffViewDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	CString formatter;

	formatter.Format("%d", pDoc->numVertices);
	theDialog.setNumberOfVertices(formatter);

	formatter.Format("%d", pDoc->numPolygons);
	theDialog.setNumberOfPolygons(formatter);

	formatter.Format("%d", pDoc->numEdges);
	theDialog.setNumberOfEdges(formatter);

	formatter.Format("%d", pDoc->numUniqueEdges);
	theDialog.setNumberOfUniqueEdges(formatter);

	formatter.Format("%d", pDoc->largestPolygonSize);
	theDialog.setLargestPolygonSize(formatter);

	formatter.Format("%d", tmanager.getNumTriangles());
	theDialog.setNumberOfTriangles(formatter);

	theDialog.DoModal();	
}

void CRoffViewView::OnToolsResetview() 
{
	SetDefaultPerspective();
	BuildStructures();
	BuildBuffer();
	Invalidate(FALSE);
}


void CRoffViewView::OnParameterDialog() 
{
	RoffParameterDialog theDialog;

	theDialog.setIlluminator(illuminator);

	if (theDialog.DoModal() == IDOK) {
		illuminator = theDialog.getIlluminator();
	    BuildStructures();
        BuildBuffer();
		Invalidate(FALSE);
	}
}

void CRoffViewView::OnViewParameters() 
{
    RoffViewParameterDialog rvpd;

	rvpd.setAt(transformer.getAt());
	rvpd.setFrom(transformer.getFrom());
	rvpd.setUp(transformer.getUp());
	rvpd.setAngle(transformer.getAngle());

	if (rvpd.DoModal() == IDOK) {
		transformer.setAt(rvpd.getAt());
		transformer.setFrom(rvpd.getFrom());
		transformer.setUpVector(rvpd.getUp());
		transformer.setAngle(rvpd.getAngle());

	    BuildStructures();
        BuildBuffer();
		Invalidate(FALSE);
	}
}

void CRoffViewView::OnViewSolid() 
{
	if (!drawPolygons) {
        drawPolygons = true;
    	drawWireframe = false;
		drawPoints = false;
        BuildStructures();
        BuildBuffer();
        Invalidate(FALSE);
	}
}

void CRoffViewView::OnUpdateViewSolid(CCmdUI* pCmdUI) 
{
   	if (drawPolygons) {
		pCmdUI->SetCheck(1);
	} else {
		pCmdUI->SetCheck(0);
	}
}

void CRoffViewView::OnViewWire() 
{
	if (!drawWireframe) {
    	drawWireframe = true;
        drawPolygons = false;
		drawPoints = false;
        BuildStructures();
        BuildBuffer();
        Invalidate(FALSE);
	}
}

void CRoffViewView::OnUpdateViewWire(CCmdUI* pCmdUI) 
{
	if (drawWireframe) {
		pCmdUI->SetCheck(1);
	} else {
		pCmdUI->SetCheck(0);
	}
}

void CRoffViewView::OnViewPoints() 
{
	if (!drawPoints) {
		drawPoints = true;
        drawPolygons = false;
    	drawWireframe = false;
        BuildStructures();
        BuildBuffer();
        Invalidate(FALSE);
	}
}

void CRoffViewView::OnUpdateViewPoints(CCmdUI* pCmdUI) 
{
	if (drawPoints) {
		pCmdUI->SetCheck(1);
	} else {
		pCmdUI->SetCheck(0);
	}
}

void CRoffViewView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	CRoffViewDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	viewMode = true;
	controlMode = false;
	lmouseDown = false;
	rmouseDown = false;
	selectedPoint = -1;

	numPoints = pDoc->numUniqueEdges * 2;
	convertedXPoints.SetSize(pDoc->numVertices);
	convertedYPoints.SetSize(pDoc->numVertices);

	tmanager.setPointArray(&(pDoc->vertices));
	tmanager.set2DPointArrays(&convertedXPoints, &convertedYPoints);
    tmanager.tessellate(&(pDoc->polygons));

	SetDefaultPerspective();
	BuildStructures();
	BuildBuffer();
	CView::OnUpdate(pSender, lHint, pHint);
}

void CRoffViewView::OnEditCopy() 
{
    if ( !OpenClipboard() ) {
        AfxMessageBox( "Cannot open the Clipboard" );
        return;
	}
    // Remove the current Clipboard contents
    if( !EmptyClipboard() ) {
	    AfxMessageBox( "Cannot empty the Clipboard" );
        return;
	}

    // Get the current picture
   	CMetaFileDC* newMetaFile = new CMetaFileDC();
	newMetaFile->CreateEnhanced(NULL, NULL, NULL, NULL);
    newMetaFile->BitBlt(upperX, upperY, squareSize, squareSize, pMemDC, 0, 0, SRCCOPY);

	HENHMETAFILE hData = newMetaFile->CloseEnhanced();
	delete newMetaFile;

    if ( ::SetClipboardData( CF_ENHMETAFILE, hData ) == NULL ) {
		DeleteEnhMetaFile(hData);
        AfxMessageBox( "Unable to set Clipboard data" );
        CloseClipboard();
        return;
	}
    CloseClipboard();
}

void CRoffViewView::OnControlEdit() 
{
	if (!controlMode) {
		controlMode = true;
        viewMode = false;
		lmouseDown = false;
		rmouseDown = false;
        BuildStructures();
        BuildBuffer();
        Invalidate(FALSE);
	}
}

void CRoffViewView::OnUpdateControlEdit(CCmdUI* pCmdUI) 
{
	if (controlMode) {
		pCmdUI->SetCheck(1);
	} else {
		pCmdUI->SetCheck(0);
	}
}

void CRoffViewView::OnControlView() 
{
	if (!viewMode) {
		viewMode = true;
        controlMode = false;
		lmouseDown = false;
		rmouseDown = false;
        BuildStructures();
        BuildBuffer();
        Invalidate(FALSE);
	}
}

void CRoffViewView::OnUpdateControlView(CCmdUI* pCmdUI) 
{
	if (viewMode) {
		pCmdUI->SetCheck(1);
	} else {
		pCmdUI->SetCheck(0);
	}
}

void CRoffViewView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	CView::OnLButtonDblClk(nFlags, point);

	CRoffViewDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// If we currently don't have a point selected, we need to
	// select one.
	if (selectedPoint == -1) {
		// Convert the click point coordinates into coordinates
		// relative to the viewing square.
	    point.x = point.x - upperX;
	    point.y = point.y - upperY;

	    int closestIndex = -1;
	    int closestDistance;
	    int distance;
	    CPoint convertedPoint;
	    RoffVector point2D;

	    BuildStructures();

	    int totalVertices = pDoc->numVertices;
    	if (totalVertices > 0) {
            closestIndex = 0;
		    convertedPoint.x = (int)(squareSize * convertedXPoints[closestIndex]);
		    convertedPoint.y = squareSize - (int)(squareSize * convertedYPoints[closestIndex]);
		    closestDistance = (((point.x - convertedPoint.x) *
    			                (point.x - convertedPoint.x)) +
					           ((point.y - convertedPoint.y) *
    					        (point.y - convertedPoint.y))
	    			          );
	        for (int i = 1;i < totalVertices;i++) {
     		    convertedPoint.x = (int)(squareSize * convertedXPoints[i]);
    	     	convertedPoint.y = squareSize - (int)(squareSize * convertedYPoints[i]);
	    		distance = (((point.x - convertedPoint.x) *
                             (point.x - convertedPoint.x)) +
			    	        ((point.y - convertedPoint.y) *
				             (point.y - convertedPoint.y))
			               );
    			if (distance < closestDistance) {
	    			closestIndex = i;
		    		closestDistance = distance;
    			}
		    }
	    }

    	selectedPoint = closestIndex;
	// If a point is currently selected, just de-select it.
	} else {
		selectedPoint = -1;
	}
	BuildBuffer();
	Invalidate(FALSE);
}
